import hmac
import random
from collections import namedtuple

Point = namedtuple('Point', ['x', 'y'])


class EllipticCurveBase:
    """ A generic class for elliptic curves and operations on them.

        The curves must be of the form: y^2 = x^3 + a*x + b.
    """

    def __init__(self, hash_function):
        self.hash_function = hash_function

    def is_on_curve(self, p):
        """ Checks whether a point is on the curve.

        Args:
            p (ECPointAffine): Point to be checked

        Returns:
            bool: True if p is on the curve, False otherwise.
        """
        raise NotImplementedError

    def y_from_x(self, x):
        """ Computes the y component corresponding to x.

            Since elliptic curves are symmetric about the x-axis,
            the x component (and sign) is all that is required to determine
            a point on the curve.

        Args:
            x (int): x component of the point.

        Returns:
            tuple: both possible y components of the point.
        """
        raise NotImplementedError

    def gen_key_pair(self, random_generator=random.SystemRandom()):
        """ Generates a public/private key pair.

        Args:
            random_generator (generator): The random generator to use.
        Returns:
            tuple: A private key in the range of 1 to `self.n - 1`
               and an ECPointAffine containing the public key point.
        """
        raise NotImplementedError

    def public_key(self, private_key):
        """ Returns the public (verifying) key for a given private key.

        Args:
            private_key (int): the private key to derive the public key for.

        Returns:
            ECPointAffine: The point representing the public key.
        """
        raise NotImplementedError

    def recover_public_key(self, message, signature, recovery_id=None):
        """ Recovers possibilities for the public key associated with the
            private key used to sign message and generate signature.

            Since there are multiple possibilities (two for curves with
            co-factor = 1), each possibility that successfully verifies the
            signature is returned.
        Args:
           message (bytes): The message that was signed.
           signature (ECPointAffine): The point representing the signature.
           recovery_id (int) (Optional): If provided, limits the valid x and y
              point to only that described by the recovery_id.

        Returns:
           list(ECPointAffine): List of points representing valid public
              keys that verify signature.
        """
        raise NotImplementedError

    def _sign(self, message, private_key, do_hash=True, secret=None):
        raise NotImplementedError

    def sign(self, message, private_key, do_hash=True):
        """ Signs a message with the given private key.

        Args:
            message (bytes): The message to be signed
            private_key (int): Integer that is the private key
            do_hash (bool): True if the message should be hashed prior
                to signing, False if not. This should always be left as
                True except in special situations which require doing
                the hash outside (e.g. handling Bitcoin bugs).

        Returns:
            (Point, int): The point (r, s) representing the signature
                and the ID representing which public key possibility
                is associated with the private key being used to sign.
        """
        return self._sign(message, private_key, do_hash)

    def verify(self, message, signature, public_key, do_hash=True):
        """ Verifies that signature was generated with a private key corresponding
            to public key, operating on message.

        Args:
            message (bytes): The message to be signed
            signature (Point): (r, s) representing the signature
            public_key (ECPointAffine): ECPointAffine of the public key
            do_hash (bool): True if the message should be hashed prior
               to signing, False if not. This should always be left as
               True except in special situations which require doing
               the hash outside (e.g. handling Bitcoin bugs).

        Returns:
            bool: True if the signature is verified, False otherwise.
        """
        raise NotImplementedError

    def _nonce_random(self):
        return random.SystemRandom().randrange(1, self.n - 1)

    def _nonce_rfc6979(self, private_key, message):
        """ Computes a deterministic nonce (k) for use when signing
            according to RFC6979 (https://tools.ietf.org/html/rfc6979),
            Section 3.2.

        Args:
            private_key (int): The private key.
            message (bytes): A hash of the input message.

        Returns:
            int: A deterministic nonce.
        """
        hash_bytes = 32
        x = private_key.to_bytes(hash_bytes, 'big')
        # Message should already be hashed by the time it gets here,
        # so don't bother doing another hash.
        x_msg = x + message

        # Step b
        V = bytes([0x1] * hash_bytes)

        # Step c
        K = bytes([0x0] * hash_bytes)

        # Step d
        K = hmac.new(K, V + bytes([0]) + x_msg, self.hash_function).digest()

        # Step e
        V = hmac.new(K, V, self.hash_function).digest()

        # Step f
        K = hmac.new(K, V + bytes([0x1]) + x_msg, self.hash_function).digest()

        # Step g
        V = hmac.new(K, V, self.hash_function).digest()

        while True:
            # Step h.1
            T = b''

            # Step h.2
            while 8 * len(T) < self.nlen:
                V = hmac.new(K, V, self.hash_function).digest()
                T += V

            # Step h.3
            k = int.from_bytes(T, 'big')
            if k >= 1 and k < (self.n - 1):
                return k

            K = hmac.new(K, V + bytes([0]), self.hash_function).digest()
            V = hmac.new(K, V, self.hash_function).digest()



from .data import (
    Account,
    BaseAccount,
    ContinuousVestingAccount,
    DelayedVestingAccount,
    PeriodicVestingAccount,
    Period,
    PublicKey,
    TxInfo,
    TxLog,
    parse_tx_logs,
    EthAccount,
)

from .msgs import (
    MsgCreatePeriodicVestingAccount,
    MsgCreateVestingAccount,
    MsgDonateAllVestingTokens
)

__all__ = [
    "Account",
    "BaseAccount",
    "ContinuousVestingAccount",
    "DelayedVestingAccount",
    "PeriodicVestingAccount",
    "Period"
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
    "MsgCreatePeriodicVestingAccount",
    "MsgCreateVestingAccount",
    "MsgDonateAllVestingTokens"
]



from __future__ import annotations

import copy
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Union

from glitter_proto.blockved.glitterchain.index import Argument as Argument_pb

from glitter_sdk.util.json import JSONSerializable

from .argument import Argument


class Arguments(JSONSerializable, List[Argument_pb]):
    """Represents an unordered collection of :class:`Argument` objects
    -- analagous to ``sdk.Arguments`` and ``sdk.DecCoins`` in Cosmos SDK. If one of the
    input Arguments would be ``Dec``-amount type Argument, the resultant Arguments is converted to
    ``Dec``-amount Arguments.

    Args:
        arg (Optional[Arguments.Input], optional): argument to convert. Defaults to ``{}``.

    Raises:
        TypeError: if ``arg`` is not an Iterable
    """

    Input = Union[Iterable[Argument],  Dict[str, Argument]]
    """Types which can be converted into a :class:`Arguments` object."""

    _arguments: Dict[str, Argument]

    def __repr__(self) -> str:
        if len(self) == 0:
            return "Arguments()"
        else:
            return f"Arguments('{self!s}')"

    def __str__(self) -> str:
        return ",".join(str(Argument) for Argument in self)

    def __init__(self, arg: Optional[Arguments.Input] = {}, **denoms):
        """Converts the argument into a :class:`Arguments` object."""

        if arg is None:
            self._arguments = {}
            return

        # arg should be an iterable
        try:
            iter(arg)
        except TypeError:
            raise TypeError(f"could not create Arguments object with argument: {arg!s}")

        if isinstance(arg, Arguments):
            self._arguments = copy.deepcopy(arg._arguments)
            return

        self._arguments = Arguments(denoms)._arguments if denoms else {}

        arguments: Iterable[Argument]
        if isinstance(arg, dict):
            arguments = [Argument(denom, arg[denom]) for denom in arg]
        else:
            arguments = arg
        for argument in arguments:
            self._arguments[argument.denom] = argument

        # make all Arguments DecCoin if one is DecCoin
        if not all([c.is_int_coin() for c in self]):
            for denom in self._arguments:
                self._arguments[denom] = self._arguments[denom].to_dec_coin()

    def __getitem__(self, denom: str) -> Argument:
        return self._arguments[denom]

    def get(self, denom: str) -> Optional[Argument]:
        """Get the Argument with the denom contained in the Arguments set.

        Args:
            denom (str): denom

        Returns:
            Optional[Argument]: result (can be ``None``)
        """
        return self._arguments.get(denom)

    @classmethod
    def from_proto(cls, proto: List[Argument_pb]) -> Arguments:
        """Converts list of Argument-data objects to :class:`Arguments`.

        Args:
            data (list): list of Argument-data objects
        """
        Arguments = map(Argument.from_proto, proto)
        return cls(Arguments)

    def to_proto(self) -> List[Argument_pb]:
        return [Argument.to_proto() for Argument in self]

    def to_list(self) -> List[Argument]:
        """Converts the set of :class:`Coin` objects contained into a sorted list by denom.

        Returns:
            List[Coin]: list, sorted by denom
        """
        return sorted(self._arguments.values(), key=lambda c: c.denom)

    def __eq__(self, other) -> bool:
        try:
            return self.to_list() == other.to_list()
        except AttributeError:
            return False

    def __iter__(self):
        return iter(self.to_list())

    def __len__(self):
        return len(self.to_list())

    def __contains__(self, denom: str) -> bool:
        return denom in self._arguments



import math
import platform
from ctypes import c_char_p
from ctypes import c_int
from ctypes import c_long
from ctypes import c_void_p
from ctypes import CDLL
from ctypes import create_string_buffer
from ctypes import POINTER
from ctypes import Structure


class OpenSSLSignature(Structure):
    _fields_ = [('r', c_void_p), ('s', c_void_p)]


sys_type = platform.system()
if sys_type == 'Darwin':
    libcrypto = CDLL('libcrypto.dylib')
elif sys_type == 'Linux':
    libcrypto = CDLL('libcrypto.so')
else:
    raise Exception('Unsupported platform %s' % sys_type)

lc = libcrypto

lc.OBJ_sn2nid.argtypes = [c_char_p]
lc.OBJ_sn2nid.restype = c_int

lc.EC_KEY_new_by_curve_name.restype = c_void_p
lc.EC_KEY_generate_key.argtypes = [c_void_p]
lc.EC_KEY_generate_key.restype = c_int
lc.EC_KEY_check_key.argtypes = [c_void_p]
lc.EC_KEY_check_key.restype = c_int
lc.EC_KEY_get0_private_key.argtypes = [c_void_p]
lc.EC_KEY_get0_private_key.restype = c_void_p
lc.EC_KEY_get0_public_key.argtypes = [c_void_p]
lc.EC_KEY_get0_public_key.restype = c_void_p
lc.EC_KEY_set_private_key.argtypes = [c_void_p, c_void_p]
lc.EC_KEY_set_private_key.restype = c_int
lc.EC_KEY_set_public_key.argtypes = [c_void_p, c_void_p]
lc.EC_KEY_set_public_key.restype = c_int
lc.EC_KEY_get0_group.argtypes = [c_void_p]
lc.EC_KEY_get0_group.restype = c_void_p
lc.EC_KEY_free.argtypes = [c_void_p]

lc.EC_GROUP_new_by_curve_name.restype = c_void_p
lc.EC_GROUP_get_order.argtypes = [c_void_p] * 3
lc.EC_GROUP_get_order.restype = c_int
lc.EC_GROUP_get_curve_GFp.argtypes = [c_void_p] * 5
lc.EC_GROUP_get_curve_GFp.restype = c_int
lc.EC_GROUP_free.argtypes = [c_void_p]

lc.EC_POINT_new.argtypes = [c_void_p]
lc.EC_POINT_new.restype = c_void_p
lc.EC_POINT_point2oct.argtypes = [c_void_p]
lc.EC_POINT_point2oct.restype = c_int
lc.EC_POINT_mul.argtypes = [c_void_p] * 6
lc.EC_POINT_mul.restype = c_int
lc.EC_POINT_invert.argtypes = [c_void_p] * 3
lc.EC_POINT_invert.restype = c_int
lc.EC_POINT_get_affine_coordinates_GFp.argtypes = [c_void_p] * 5
lc.EC_POINT_set_affine_coordinates_GFp.argtypes = [c_void_p] * 5
lc.EC_POINT_set_affine_coordinates_GFp.restype = c_int
lc.EC_POINT_set_compressed_coordinates_GFp.argtypes = [c_void_p, c_void_p, c_void_p, c_int, c_void_p]
lc.EC_POINT_set_compressed_coordinates_GFp.restype = c_int
lc.EC_POINT_is_at_infinity.argtypes = [c_void_p, c_void_p]
lc.EC_POINT_is_at_infinity.restype = c_int
lc.EC_POINT_set_to_infinity.argtypes = [c_void_p, c_void_p]
lc.EC_POINT_set_to_infinity.restype = c_int
lc.EC_POINT_is_on_curve.argtypes = [c_void_p] * 3
lc.EC_POINT_is_on_curve.restype = c_int

lc.d2i_ECDSA_SIG.argtypes = [c_void_p, c_void_p, c_long]
lc.d2i_ECDSA_SIG.restype = POINTER(OpenSSLSignature)
lc.ECDSA_SIG_new.argtypes = []
lc.ECDSA_SIG_new.restype = POINTER(OpenSSLSignature)
lc.ECDSA_SIG_free.argtypes = [POINTER(OpenSSLSignature)]
lc.ECDSA_do_sign_ex.argtypes = [c_char_p, c_int, c_void_p, c_void_p, c_void_p]
lc.ECDSA_do_sign_ex.restype = POINTER(OpenSSLSignature)
lc.ECDSA_do_verify.argtypes = [c_void_p, c_int, POINTER(OpenSSLSignature), c_void_p]
lc.ECDSA_do_verify.restype = c_int

lc.BN_new.restype = c_void_p
lc.BN_num_bits.argtypes = [c_void_p]
lc.BN_num_bits.restype = c_int
lc.BN_bn2bin.argtypes = [c_void_p, c_char_p]
lc.BN_bn2bin.restype = c_int
lc.BN_bin2bn.argtypes = [c_char_p, c_int, c_void_p]
lc.BN_bin2bn.restype = c_void_p
lc.BN_mod_mul.argtypes = [c_void_p] * 5
lc.BN_mod_mul.restype = c_int
lc.BN_mod_add.argtypes = [c_void_p] * 5
lc.BN_mod_add.restype = c_int
lc.BN_mod_inverse.argtypes = [c_void_p] * 4
lc.BN_mod_inverse.restype = c_void_p
lc.BN_free.argtypes = [c_void_p]
lc.BN_CTX_new.restype = c_void_p
lc.BN_CTX_start.argtypes = [c_void_p]
lc.BN_CTX_get.argtypes = [c_void_p]
lc.BN_CTX_get.restype = c_void_p
lc.BN_CTX_end.argtypes = [c_void_p]
lc.BN_CTX_free.argtypes = [c_void_p]

lc.ERR_load_crypto_strings()


def get_curve_params(group):
    """ Retrieves all elliptic curve parameters

    Args:
        group (c_void_p): The OpenSSL group (curve) to
            retrieve the parameters for.

    Returns:
        dict: With the following key/value pairs:
            p: Prime defining the field
            a: Linear coefficient of the curve
            b: Curve constant
            n: Curve order
            h: Curve co-factor
    """
    ctx = c_void_p(lc.BN_CTX_new())
    lc.BN_CTX_start(ctx)

    order_bn = c_void_p(lc.BN_CTX_get(ctx))
    cofactor_bn = c_void_p(lc.BN_CTX_get(ctx))
    p_bn = c_void_p(lc.BN_CTX_get(ctx))
    a_bn = c_void_p(lc.BN_CTX_get(ctx))
    b_bn = c_void_p(lc.BN_CTX_get(ctx))
    lc.EC_GROUP_get_order(group, order_bn, c_void_p())
    lc.EC_GROUP_get_cofactor(group, cofactor_bn, c_void_p())
    lc.EC_GROUP_get_curve_GFp(group, p_bn, a_bn, b_bn, ctx)

    rv = {}
    rv['p'] = bn_to_int(p_bn)
    rv['a'] = bn_to_int(a_bn)
    rv['b'] = bn_to_int(b_bn)
    rv['n'] = bn_to_int(order_bn)
    rv['h'] = bn_to_int(cofactor_bn)

    lc.BN_CTX_end(ctx)
    lc.BN_CTX_free(ctx)

    return rv


def new_key(curve_name, private_key=None):
    """ Generates a new EC_KEY

    Args:
        curve_name (int): The OpenSSL identifier of the curve which
            the key will be part of.
        private_key (int): If not provided, a random key pair will be
            generated. Otherwise, the key will be initiated with the
            provided private key (and corresponding public key)

    Returns:
        c_void_p: An opaque pointer to the new EC_KEY object. The
            caller is responsible for freeing the key object.
    """
    k = c_void_p(lc.EC_KEY_new_by_curve_name(curve_name))

    if private_key is None:
        lc.EC_KEY_generate_key(k)
    else:
        key_ok = set_private_key_from_int(k, private_key)
        if not key_ok:
            raise ValueError('Key is not ok')

    return k


def get_private_key_bytes(key):
    """ Retrieves the bytes corresponding to the private key.

    Args:
        key (c_void_p): A pointer to an EC_KEY object.

    Returns:
        bytes: A byte stream containing the private key.
    """
    priv_key_bn = c_void_p(lc.EC_KEY_get0_private_key(key))
    priv_key_bytes = bn_to_bytes(priv_key_bn)

    return priv_key_bytes


def get_private_key_int(key):
    """ Retrieves the integer corresponding to the private key.

    Args:
        key (c_void_p): A pointer to an EC_KEY object.

    Returns:
        int: The integer representation of the private key.
    """
    rv = None
    b = get_private_key_bytes(key)
    if b is not None:
        rv = int.from_bytes(b, byteorder='big')

    return rv


def set_private_key_from_bytes(key, b):
    """ Sets the private portion of key from b.

        This function also sets the public key portion of
        key after computing the public key from the provided
        bytes of the private key.

    Args:
        key (c_void_p): A pointer to an EC_KEY object.
        b (bytes): The bytes representing the private key.

    Returns:
        bool: Whether key passes OpenSSL's sanity checks after
            both private & public keys are set.
    """
    priv_bn = bytes_to_bn(b)

    group = c_void_p(lc.EC_KEY_get0_group(key))

    lc.EC_KEY_set_private_key(key, priv_bn)

    # Now jam the public key
    pub_pt = c_void_p(lc.EC_POINT_new(group))
    lc.EC_POINT_mul(group, pub_pt, priv_bn, c_void_p(), c_void_p(), c_void_p())
    lc.EC_KEY_set_public_key(key, pub_pt)

    lc.EC_POINT_free(pub_pt)

    return lc.EC_KEY_check_key(key)


def set_private_key_from_int(key, i, size=32):
    """ Sets the private portion of key from i.

        This function also sets the public key portion of
        key after computing the public key from the provided
        private key.

    Args:
        key (c_void_p): A pointer to an EC_KEY object.
        i (int): The integer representing the private key.
        size (int): Size in bytes the key should be.

    Returns:
        bool: Whether key passes OpenSSL's sanity checks after
            both private & public keys are set.
    """
    b = i.to_bytes(size, byteorder='big')
    return set_private_key_from_bytes(key, b)


def bn_to_bytes(bn):
    """ Translates an OpenSSL big number (BN) to bytes.

    Args:
        bn (c_void_p): A pointer to an OpenSSL BN object.

    Returns:
        bytes: The bytes corresponding to the BN object. It is
            returned as a big-endian positive number.
    """
    size = math.ceil(lc.BN_num_bits(bn) / 8.0)
    b = create_string_buffer(b'\x00' * size)
    lc.BN_bn2bin(bn, b)

    return b[:size]


def bytes_to_bn(b, bn=None):
    """ Translates bytes to an OpenSSL big number (BN).

    Args:
        b (bytes): Big-endian, positive number to translate.
        bn (c_void_p): If not None, sets it to the number represented
            by b.

    Returns:
        c_void_p: If bn is None, a pointer to a new OpenSSL BN object,
             otherwise bn. In all cases, the caller is responsible for
             freeing the memory associated with the OpenSSL object.
    """
    buf = create_string_buffer(b)
    return c_void_p(lc.BN_bin2bn(buf, len(b), bn))


def bn_to_int(bn):
    """ Translates an OpenSSL big number (BN) to a Python integer.

    Args:
        bn (c_void_p): A pointer to an OpenSSL BN object.

    Returns:
        int: The python integer corresponding to the BN object.
    """
    return int.from_bytes(bn_to_bytes(bn), byteorder='big')


def int_to_bn(i, bn=None, size=32):
    """ Translates a python integer to an OpenSSL big number (BN).

    Args:
        i (int): Integer to translate.
        bn (c_void_p): If not None, sets it to the number represented
            by i.
        size (int): The maximal byte-length of i.

    Returns:
        c_void_p: If bn is None, a pointer to a new OpenSSL BN object,
             otherwise bn. In all cases, the caller is responsible for
             freeing the memory associated with the OpenSSL object.
    """
    return bytes_to_bn(i.to_bytes(size, byteorder='big'), bn)


def point_get_xy_bytes(group, pt):
    """ Gets bytes for x and y components of an EC_POINT.

        This uses `bn_to_bytes` to create the bytes objects after
        extracting the x and y components of pt.

    Args:
        group (c_void_p): An opaque pointer to the group (curve) that
            pt is part of.
        pt (c_void_p): An opaque pointer to the OpenSSL EC_POINT object.

    Returns:
        tuple: Containing the x bytes, y bytes and a boolean representing
            whether the point is at infinity or not.
    """
    x_bn = c_void_p(lc.BN_new())
    y_bn = c_void_p(lc.BN_new())
    lc.EC_POINT_make_affine(group, pt, c_void_p(None))
    lc.EC_POINT_get_affine_coordinates_GFp(group, pt, x_bn, y_bn, c_void_p(None))
    inf = bool(lc.EC_POINT_is_at_infinity(group, pt))

    x_bytes = bn_to_bytes(x_bn)
    y_bytes = bn_to_bytes(y_bn)

    lc.BN_free(x_bn)
    lc.BN_free(y_bn)

    return (x_bytes, y_bytes, inf)


def point_get_xy_ints(group, pt):
    """ Gets the x and y components of an EC_POINT.

    Args:
        group (c_void_p): An opaque pointer to the group (curve) that
            pt is part of.
        pt (c_void_p): An opaque pointer to the OpenSSL EC_POINT object.

    Returns:
        tuple: Containing x, y and a boolean representing whether the
            point is at infinity or not.
    """
    x_b, y_b, inf = point_get_xy_bytes(group, pt)
    x = int.from_bytes(x_b, byteorder='big')
    y = int.from_bytes(y_b, byteorder='big')

    return (x, y, inf)


def point_new_from_bytes(group, x_bytes, y_bytes, infinity=False):
    """ Creates a new OpenSSL EC_POINT from bytes.

    Args:
        group (c_void_p): An opaque pointer to the group (curve) that
            the point should be part of.
        x_bytes (bytes): Big-endian, positive byte representation of x.
        y_bytes (bytes): Big-endian, positive byte representation of y.
        infinity (bool): True if the point is at infinity, False otherwise.

    Returns:
        c_void_p: An opaque pointer to the newly constructed OpenSSL
            EC_POINT object. The caller bears responsibility for freeing
            the memory associated with the returned object.
    """
    pt = c_void_p(lc.EC_POINT_new(group))

    x_bn = bytes_to_bn(x_bytes)
    y_bn = bytes_to_bn(y_bytes)

    res = lc.EC_POINT_set_affine_coordinates_GFp(group, pt, x_bn, y_bn, c_void_p(None))
    if not res:
        lc.EC_POINT_free(pt)
        return None

    if infinity:
        lc.EC_POINT_set_to_infinity(group, pt)

    return pt


def point_new_from_ints(group, x, y, infinity=False, size=32):
    """ Creates a new OpenSSL EC_POINT from x & y integers.

    Args:
        group (c_void_p): An opaque pointer to the group (curve) that
            the point should be part of.
        x_bytes (bytes): Big-endian, positive byte representation of x.
        y_bytes (bytes): Big-endian, positive byte representation of y.
        infinity (bool): True if the point is at infinity, False otherwise.
        size (int): Maximal byte-length of x and y.

    Returns:
        c_void_p: An opaque pointer to the newly constructed OpenSSL
            EC_POINT object. The caller bears responsibility for freeing
            the memory associated with the returned object.
    """
    x_bytes = x.to_bytes(size, byteorder='big')
    y_bytes = y.to_bytes(size, byteorder='big')

    return point_new_from_bytes(group, x_bytes, y_bytes, infinity)


def get_public_key_bytes(key):
    """ Returns the bytes corresponding to the public key coordinates.

    Args:
        key (c_void_p): An opaque pointer to an OpenSSL EC_KEY object.

    Returns:
        tuple: Containing bytes for x & y public key components and a
            boolean indicating whether the point is at infinity or not.
    """
    group = c_void_p(lc.EC_KEY_get0_group(key))
    pub_key_point = c_void_p(lc.EC_KEY_get0_public_key(key))
    x_b, y_b, inf = point_get_xy_bytes(group, pub_key_point)

    return (x_b, y_b, inf)


def get_public_key_ints(key):
    """ Returns the integers corresponding to the public key coordinates.

    Args:
        key (c_void_p): An opaque pointer to an OpenSSL EC_KEY object.

    Returns:
        tuple: Containing the x & y public key components and a
            boolean indicating whether the point is at infinity or not.
    """
    xb, yb, inf = get_public_key_bytes(key)
    x = int.from_bytes(xb, byteorder='big')
    y = int.from_bytes(yb, byteorder='big')

    return (x, y, inf)


def set_public_key_from_bytes(key, x_bytes, y_bytes, infinity):
    """ Sets the public key portion of an OpenSSL EC_KEY object from
        bytes representations of the x & y components of the public key.

    Args:
        key (c_void_p): An opaque pointer to an OpenSSL EC_KEY object.
        x_bytes (bytes): Big-endian, positive byte-representation of the
            x component of the public-key point.
        y_bytes (bytes): Big-endian, positive byte-representation of the
            y component of the public-key point.
        infinity (bool): True if the point is at infinity, False otherwise.

    Returns:
        bool: True if the public key was set properly, False otherwise.
    """
    group = c_void_p(lc.EC_KEY_get0_group(key))
    pub_pt = point_new_from_bytes(group, x_bytes, y_bytes, infinity)
    res = lc.EC_KEY_set_public_key(key, pub_pt)

    return bool(res)


def set_public_key_from_ints(key, x, y, infinity, size=32):
    """ Sets the public key portion of an OpenSSL EC_KEY object from
        bytes representations of the x & y components of the public key.

    Args:
        key (c_void_p): An opaque pointer to an OpenSSL EC_KEY object.
        x (int): X component of the public-key point.
        y (int): Y component of the public-key point.
        infinity (bool): True if the point is at infinity, False otherwise.
        size (int): Maximal byte-size of x and y.

    Returns:
        bool: True if the public key was set properly, False otherwise.
    """
    x_bytes = x.to_bytes(size, byteorder='big')
    y_bytes = y.to_bytes(size, byteorder='big')
    return set_public_key_from_bytes(key, x_bytes, y_bytes, infinity)


def sig_new_from_bytes(r_bytes, s_bytes):
    """ Creates a new OpenSSL ECDSA_SIG structure from r & s bytes.

    Args:
        r_bytes (bytes): Big-endian, positive byte-representation of the
            r component of the signature.
        s_bytes (bytes): Big-endian, positive byte-representation of the
            s component of the signature.

    Returns:
        c_void_p: An opaque pointer to a new OpenSSL ECDSA_SIG structure
            represented by the `OpenSSLSignature` class.
    """
    sig = lc.ECDSA_SIG_new()

    r_buf = create_string_buffer(r_bytes)
    s_buf = create_string_buffer(s_bytes)
    lc.BN_bin2bn(r_buf, len(r_bytes), sig.contents.r)
    lc.BN_bin2bn(s_buf, len(s_bytes), sig.contents.s)

    return sig


def sig_new_from_ints(r, s, size=32):
    """ Creates a new OpenSSL ECDSA_SIG structure from r & s integers.

    Args:
        r (int): R component of the signature.
        s (int): S component of the signature.
        size (int): Maximal byte-size of r & s.

    Returns:
        c_void_p: An opaque pointer to a new OpenSSL ECDSA_SIG structure
            represented by the `OpenSSLSignature` class.
    """
    return sig_new_from_bytes(r_bytes=r.to_bytes(size, byteorder='big'), s_bytes=s.to_bytes(size, byteorder='big'))



def highlight_prepare(fields: list[str]):
    place_hold = ",".join(["\"{}\""] * len(fields))
    option = """{"highlight":{ "style":"html","fields":[""" + place_hold.format(*fields) + """]}}"""
    option = option.translate({ord('"'): "\""})
    # return "{}{}{}".format("/*+ SET_VAR( bleve_option=", option, ")*/ ")
    return "/*+ SET_VAR(full_text_option='%s')*/" % option



"""Auth mdoule message types."""

from __future__ import annotations

from glitter_proto.cosmos.vesting.v1beta1 import (
    MsgDonateAllVestingTokens as MsgDonateAllVestingTokens_pb,
)

from glitter_sdk.core import AccAddress, Coins
from glitter_sdk.core.msg import Msg

__all__ =["MsgDonateAllVestingTokens"]

import attr

@attr.s
class MsgDonateAllVestingTokens(Msg):
    """MsgDonateAllVestingTokens defines a message that enables donating all vesting.
    
    Args:
        from_address (AccAddress): address donating all vesting tokens.
    """

    type_amino = "cosmos-sdk/MsgDonateAllVestingTokens"
    type_url = "/cosmos.vesting.v1beta1.MsgDonateAllVestingTokens"
    prototype = MsgDonateAllVestingTokens_pb

    from_address: AccAddress = attr.ib()

    def to_amino(self) -> dict :
        return {
            "type": self.type_amino,
            "value": {
                "from_address": self.from_address,
            },
        }

    @classmethod
    def from_data(cls, data:dict) -> MsgDonateAllVestingTokens:
        return cls(
            from_address=data["from_address"],
        )
    
    def to_data(self) -> dict :
        return {
            "@type": self.type_url,
            "from_address": self.from_address,
        }

    @classmethod
    def from_proto(cls, proto : MsgDonateAllVestingTokens_pb) -> MsgDonateAllVestingTokens:
        return cls(
            from_address = proto.from_address,
        )
    
    def to_proto(self) -> MsgDonateAllVestingTokens_pb :
        proto = MsgDonateAllVestingTokens_pb()
        proto.from_address = self.from_address
        return proto



from glitter_sdk.util.encrypt.crypto.ecdsa import ECPointAffine  # NOQA: F401
from glitter_sdk.util.encrypt.crypto.ecdsa import EllipticCurve  # NOQA: F401
from glitter_sdk.util.encrypt.crypto.ecdsa import secp256k1  # NOQA: F401



"""Data objects pertaining to accounts."""

from __future__ import annotations

from typing import Optional

import attr
from glitter_proto.ethermint.types.v1 import EthAccount as EthAccount_pb

from ....core import AccAddress
from ....util.json import JSONSerializable
from ...public_key import PublicKey
from .base_account import BaseAccount

__all__ = ["EthAccount"]


@attr.s
class EthAccount(JSONSerializable):
    """
    Stores information about an account.

    Args:
        base_account (BaseAccount): base_account
        code_hash (str): code_hash
    """

    type_amino = "cosmos-sdk/BaseAccount"
    type_url = "/ethermint.types.v1.EthAccount"

    # base_account: BaseAccount = attr.ib(converter=BaseAccount, default=None)
    base_account: BaseAccount = attr.ib()
    """"""
    code_hash: str = attr.ib(converter=str, default=None)
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "base_account": self.base_account,
                "code_hash": str(self.code_hash),
            },
        }

    @classmethod
    def from_amino(cls, amino: dict) -> EthAccount:
        amino = amino["value"] if "value" in amino else amino
        return cls(
            base_account=amino["base_account"],
            code_hash=amino["code_hash"],
        )

    def get_account_number(self) -> int:
        return self.base_account.account_number

    def get_sequence(self) -> int:
        return self.base_account.sequence

    def get_public_key(self) -> PublicKey:
        return self.base_account.public_key

    def to_data(self) -> dict:
        return {
            "base_account": self.base_account.to_data(),
            "code_hash": str(self.code_hash),
        }

    @classmethod
    def from_data(cls, data: dict) -> EthAccount:
        return cls(
            base_account=BaseAccount.from_data(data.get("base_account")),
            code_hash=data.get("code_hash")
        )

    @classmethod
    def from_proto(cls, proto: EthAccount_pb) -> EthAccount:
        return cls(
            base_account=proto.base_account,
            code_hash=proto.code_hash,
        )

    def to_proto(self) -> EthAccount_pb:
        return EthAccount_pb(
            base_account=self.base_account,
            code_hash=self.code_hash,
        )



import hashlib
import math
import random
from ctypes import c_char_p
from ctypes import c_void_p
from ctypes import create_string_buffer

from evmospy.evmoswallet.crypto import openssl as ossl
from evmospy.evmoswallet.crypto.ecdsa_base import EllipticCurveBase
from evmospy.evmoswallet.crypto.ecdsa_base import Point


class ECPointAffine:
    """ An affine (2D) representation of an elliptic curve point.

    In this implementation, only the minimum functionality
    required for bitcoin crypto-API compatibility is provided.
    All math operations make use of OpenSSL primitives.

    Args:
        curve (EllipticCurve): The curve the point is on.
        x (int): x component of point.
        y (int): y component of point.
        infinity (bool): Whether or not this point is at infinity.

    Returns:
        ECPointAffine: the point formed by (x, y) on curve.
    """

    def __init__(self, curve, x, y, infinity=False):
        self.x = x
        self.y = y
        self.curve = curve
        self.infinity = infinity

    def __str__(self):
        return f'(0x{self.x:x}, 0x{self.y:x})'

    def __eq__(self, b):
        return ((self.x == b.x) and (self.y == b.y)) or \
            (self.infinity and b.infinity)

    def __add__(self, b):
        assert self.curve == b.curve

        a_pt = ossl.point_new_from_ints(self.curve.os_group, self.x, self.y, self.infinity)
        b_pt = ossl.point_new_from_ints(b.curve.os_group, b.x, b.y, b.infinity)
        ossl.lc.EC_POINT_add(self.curve.os_group, a_pt, a_pt, b_pt, None)

        x, y, inf = ossl.point_get_xy_ints(self.curve.os_group, a_pt)

        ossl.lc.EC_POINT_free(a_pt)
        ossl.lc.EC_POINT_free(b_pt)

        return ECPointAffine(self.curve, x, y, inf)

    @property
    def compressed_bytes(self):
        """ Returns the compressed bytes for this point.

        If pt.y is odd, 0x03 is pre-pended to pt.x.
        If pt.y is even, 0x02 is pre-pended to pt.x.

        Returns:
            bytes: Compressed byte representation.
        """
        nbytes = math.ceil(self.curve.nlen / 8)
        return bytes([(self.y & 0x1) + 0x02]) + self.x.to_bytes(nbytes, 'big')

    def __bytes__(self):
        """ Returns the full-uncompressed point
        """
        nbytes = math.ceil(self.curve.nlen / 8)
        return bytes([0x04]) + self.x.to_bytes(nbytes, 'big') + self.y.to_bytes(nbytes, 'big')


class EllipticCurve(EllipticCurveBase):
    """ A generic class for elliptic curves and operations on them.

    The curves must be of the form: y^2 = x^3 + a*x + b.

    Args:
        hash_function (function): The function to use for hashing messages.
    """
    curve_name = None

    def __init__(self, hash_function):
        super().__init__(hash_function)
        self.os_group = c_void_p(ossl.lc.EC_GROUP_new_by_curve_name(self.curve_name))

        params = ossl.get_curve_params(self.os_group)

        self.p = params['p']
        self.a = params['a']
        self.b = params['b']
        self.n = params['n']
        self.h = params['h']

        self.nlen = self.n.bit_length()
        self.plen = self.p.bit_length()

        # We keep a pointer to the libcrypto CDLL object so that
        # it is guaranteed to be around when __del__ is called. If
        # we don't, a race condition exists whereby if the garbage
        # collector disposes of ossl.lc before this object, we won't
        # be able to free self.os_group (in __del__).
        self._lc = ossl.lc

    def __del__(self):
        self._lc.EC_GROUP_free(self.os_group)

    def is_on_curve(self, p):
        """ Checks whether a point is on the curve.

        Args:
            p (ECPointAffine): Point to be checked

        Returns:
            bool: True if p is on the curve, False otherwise.
        """
        ec_pt = ossl.point_new_from_ints(self.os_group, p.x, p.y)
        on_curve = ossl.lc.EC_POINT_is_on_curve(self.os_group, ec_pt, None)
        ossl.lc.EC_POINT_free(ec_pt)

        return bool(on_curve)

    def y_from_x(self, x):
        """ Computes the y component corresponding to x.

        Since elliptic curves are symmetric about the x-axis,
        the x component (and sign) is all that is required to determine
        a point on the curve.

        Args:
            x (int): x component of the point.

        Returns:
            tuple: both possible y components of the point.
        """
        rv = []
        x_bn = ossl.int_to_bn(x)
        for y_bit in [0, 1]:
            # Create a new point
            ec_pt = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
            ossl.lc.EC_POINT_set_compressed_coordinates_GFp(self.os_group, ec_pt, x_bn, y_bit, c_void_p())

            on_curve = ossl.lc.EC_POINT_is_on_curve(self.os_group, ec_pt, c_void_p())
            if not on_curve:
                ossl.lc.EC_POINT_free(ec_pt)
                rv.append(None)
                continue

            # Get the y value
            _, y, _ = ossl.point_get_xy_ints(self.os_group, ec_pt)
            rv.append(y)
            ossl.lc.EC_POINT_free(ec_pt)

        ossl.lc.BN_free(x_bn)

        return rv

    def gen_key_pair(self, random_generator=random.SystemRandom()):
        """ Generates a public/private key pair.

        Args:
            random_generator (generator): The random generator to use.

        Returns:
            tuple:
                A private key in the range of 1 to `self.n - 1`
                and an ECPointAffine containing the public key point.
        """
        private = random_generator.randrange(1, self.n)
        return private, self.public_key(private)

    def public_key(self, private_key):
        """ Returns the public (verifying) key for a given private key.

        Args:
            private_key (int): the private key to derive the public key for.

        Returns:
            ECPointAffine: The point representing the public key.
        """
        k = ossl.new_key(self.curve_name, private_key)
        pub_x, pub_y, is_inf = ossl.get_public_key_ints(k)
        ossl.lc.EC_KEY_free(k)
        return ECPointAffine(self, pub_x, pub_y, is_inf)

    def recover_public_key(self, message, signature, recovery_id=None):
        """ Recovers possibilities for the public key associated with the
        private key used to sign message and generate signature.

        Since there are multiple possibilities (two for curves with
        co-factor = 1), each possibility that successfully verifies the
        signature is returned.

        Args:
           message (bytes): The message that was signed.
           signature (ECPointAffine): The point representing the signature.
           recovery_id (int) (Optional): If provided, limits the valid x and y
              point to only that described by the recovery_id.

        Returns:
           list(ECPointAffine): List of points representing valid public
           keys that verify signature.
        """
        r = signature.x
        s = signature.y

        ctx = c_void_p(ossl.lc.BN_CTX_new())
        ossl.lc.BN_CTX_start(ctx)

        order_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        x_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        i_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        in_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        p_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        r_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        s_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        rinv_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        z_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        ossl.lc.EC_GROUP_get_order(self.os_group, order_bn, ctx)

        ossl.int_to_bn(self.p, p_bn)
        ossl.int_to_bn(r, r_bn)
        ossl.int_to_bn(s, s_bn)
        ossl.lc.BN_mod_inverse(rinv_bn, r_bn, order_bn, ctx)

        if recovery_id is not None:
            i_list = [recovery_id >> 1]
            k_list = [recovery_id & 0x1]
        else:
            i_list = range(2)
            k_list = range(2)

        rv = []
        num_bytes = math.ceil(self.nlen / 8)

        z = int.from_bytes(self.hash_function(message).digest()[:num_bytes], 'big')
        ossl.int_to_bn(z, z_bn)

        zG = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
        sR = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
        temp = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
        pub_key = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
        Rn = c_void_p(ossl.lc.EC_POINT_new(self.os_group))

        for i in i_list:
            ossl.int_to_bn(i, i_bn)
            ossl.lc.BN_mod_mul(in_bn, i_bn, order_bn, p_bn, ctx)
            ossl.lc.BN_mod_add(x_bn, r_bn, in_bn, p_bn, ctx)
            x = ossl.bn_to_int(x_bn)
            ys = self.y_from_x(x)

            for k in k_list:
                y = ys[k]
                if y & 0x1 != k:
                    y = ys[k ^ 1]

                R = ossl.point_new_from_ints(self.os_group, r, y)
                ossl.lc.EC_POINT_mul(self.os_group, Rn, None, R, order_bn, ctx)
                if not ossl.lc.EC_POINT_is_at_infinity(self.os_group, Rn):
                    continue

                ossl.lc.EC_POINT_mul(self.os_group, zG, z_bn, None, None, ctx)
                ossl.lc.EC_POINT_invert(self.os_group, zG, ctx)
                ossl.lc.EC_POINT_mul(self.os_group, sR, None, R, s_bn, ctx)
                ossl.lc.EC_POINT_add(self.os_group, temp, sR, zG, ctx)
                ossl.lc.EC_POINT_mul(self.os_group, pub_key, None, temp, rinv_bn, ctx)

                ossl.lc.EC_POINT_free(R)

                # Convert to ECPointAffine
                pub_x, pub_y, inf = ossl.point_get_xy_ints(self.os_group, pub_key)
                rv.append((ECPointAffine(self, pub_x, pub_y, inf), 2 * i + k))

        ossl.lc.EC_POINT_free(zG)
        ossl.lc.EC_POINT_free(sR)
        ossl.lc.EC_POINT_free(temp)
        ossl.lc.EC_POINT_free(pub_key)
        ossl.lc.EC_POINT_free(Rn)

        ossl.lc.BN_CTX_end(ctx)
        ossl.lc.BN_CTX_free(ctx)

        return rv

    def _sign(self, message, private_key, do_hash=True, secret=None):
        # This function computes k, kinv and rp before sending into
        # OpenSSL ECDSA_do_sign_ex() rather than using ECDSA_do_sign()
        # directly. The reason for this is that as of the commit that
        # introduces this comment, OpenSSL only supports deterministic
        # signing nonces (RFC6979) in the master branch and not any
        # release. As a result, we cannot depend on callers having any
        # OpenSSL version capable of RFC6979 deterministic nonces.
        # Nevertheless, computation of kinv (from k) and rp is done
        # using OpenSSL primitives.
        ossl.lc.ERR_clear_error()

        hashed = self.hash_function(message).digest() if do_hash else message

        r = 0
        s = 0
        recovery_id = 0

        key = ossl.new_key(self.curve_name, private_key)

        ctx = c_void_p(ossl.lc.BN_CTX_new())
        ossl.lc.BN_CTX_start(ctx)

        order_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        k_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        kinv_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        px_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        r_bn = c_void_p(ossl.lc.BN_CTX_get(ctx))
        ossl.lc.EC_GROUP_get_order(self.os_group, order_bn, ctx)

        while r == 0 or s == 0:
            k = self._nonce_rfc6979(private_key, hashed) if secret is None else secret
            ossl.int_to_bn(k, k_bn)

            ossl.lc.BN_mod_inverse(kinv_bn, k_bn, order_bn, ctx)

            p = c_void_p(ossl.lc.EC_POINT_new(self.os_group))
            ossl.lc.EC_POINT_mul(self.os_group, p, k_bn, c_void_p(), c_void_p(), ctx)
            assert self.h == 1

            px, py, _ = ossl.point_get_xy_ints(self.os_group, p)
            recovery_id = 2 if px > self.n else 0
            recovery_id |= (py & 0x1)

            # Get r
            ossl.int_to_bn(px, px_bn)
            ossl.lc.BN_nnmod(r_bn, px_bn, order_bn, ctx)
            r = ossl.bn_to_int(r_bn)

            if r == 0:
                continue

            hashed_buf = c_char_p(hashed)
            sig = ossl.lc.ECDSA_do_sign_ex(hashed_buf, len(hashed), kinv_bn, r_bn, key)
            err = ossl.lc.ERR_peek_error()
            if err:
                err_buf = create_string_buffer(120)
                ossl.lc.ERR_error_string(err, err_buf)
                raise Exception('Problem when signing: %s' % err_buf.raw.decode())

            sig_r = ossl.bn_to_int(sig.contents.r)
            sig_s = ossl.bn_to_int(sig.contents.s)

            if sig_r != r:
                raise ValueError("Didn't get the same r value.")
            s = sig_s

        ossl.lc.EC_KEY_free(key)
        ossl.lc.BN_CTX_end(ctx)
        ossl.lc.BN_CTX_free(ctx)

        return (Point(r, s), recovery_id)

    def verify(self, message, signature, public_key, do_hash=True):
        """ Verifies that signature was generated with a private key corresponding
        to public key, operating on message.

        Args:
            message (bytes): The message to be signed
            signature (Point): (r, s) representing the signature
            public_key (ECPointAffine): ECPointAffine of the public key
            do_hash (bool): True if the message should be hashed prior
               to signing, False if not. This should always be left as
               True except in special situations which require doing
               the hash outside (e.g. handling Bitcoin bugs).

        Returns:
            bool: True if the signature is verified, False otherwise.
        """
        r = signature.x
        s = signature.y

        sig = ossl.sig_new_from_ints(r, s)

        hashed = self.hash_function(message).digest() if do_hash else message

        key = c_void_p(ossl.lc.EC_KEY_new_by_curve_name(self.curve_name))
        ossl.set_public_key_from_ints(key=key, x=public_key.x, y=public_key.y, infinity=public_key.infinity)

        dig_buf = create_string_buffer(hashed)
        verified = ossl.lc.ECDSA_do_verify(dig_buf, len(hashed), sig, key)

        ossl.lc.ECDSA_SIG_free(sig)
        ossl.lc.EC_KEY_free(key)

        return bool(verified)


class p256(EllipticCurve):
    curve_name = ossl.lc.OBJ_sn2nid(c_char_p(b'prime256v1'))

    def __init__(self):
        super().__init__(hashlib.sha256)


class secp256k1(EllipticCurve):
    curve_name = ossl.lc.OBJ_sn2nid(c_char_p(b'secp256k1'))

    def __init__(self):
        super().__init__(hashlib.sha256)



from __future__ import annotations

import math
import re
from typing import Union

import attr
from glitter_proto.blockved.glitterchain.index import Argument as Argument_pb
from glitter_proto.blockved.glitterchain.index import ArgumentVarType as ArgumentVarType_pb

from glitter_sdk.util.json import JSONSerializable



@attr.s(frozen=True)
class Argument(JSONSerializable):
    """Represents a (type, value) pairing, analagous to ``sdk.Argument`` and ``sdk.DecCoin``
    in Cosmos SDK. Used for representing Glitter native assets.
    """
    type: ArgumentVarType_pb = attr.ib()
    """Argument's denomination, only ``uluna``."""

    value: bytes = attr.ib()  # type: ignore
    """Argument's value -- can be a ``int`` or :class:`Dec`"""

    @classmethod
    def from_proto(cls, proto: Argument_pb) -> Argument:
        return cls(proto.type, proto.value)

    def to_proto(self) -> Argument_pb:
        Argument = Argument_pb()
        Argument.type = self.type
        Argument.value = str(self.value)
        return Argument



import binascii, hashlib, hmac, struct
from ecdsa.curves import SECP256k1
from eth_utils import to_checksum_address, keccak as eth_utils_keccak

BIP39_PBKDF2_ROUNDS = 2048
BIP39_SALT_MODIFIER = "mnemonic"
BIP32_PRIVDEV = 0x80000000
BIP32_CURVE = SECP256k1
BIP32_SEED_MODIFIER = b'Bitcoin seed'
ETH_DERIVATION_PATH = "m/44'/60'/0'/0"


class PublicKey:
    def __init__(self, private_key):
        self.point = int.from_bytes(private_key, byteorder='big') * BIP32_CURVE.generator

    def __bytes__(self):
        xstr = self.point.x().to_bytes(32, byteorder='big')
        parity = self.point.y() & 1
        return (2 + parity).to_bytes(1, byteorder='big') + xstr

    def address(self):
        x = self.point.x()
        y = self.point.y()
        s = x.to_bytes(32, 'big') + y.to_bytes(32, 'big')
        return to_checksum_address(eth_utils_keccak(s)[12:])


def mnemonic_to_bip39seed(mnemonic, passphrase):
    mnemonic = bytes(mnemonic, 'utf8')
    salt = bytes(BIP39_SALT_MODIFIER + passphrase, 'utf8')
    return hashlib.pbkdf2_hmac('sha512', mnemonic, salt, BIP39_PBKDF2_ROUNDS)


def bip39seed_to_bip32masternode(seed):
    k = seed
    h = hmac.new(BIP32_SEED_MODIFIER, seed, hashlib.sha512).digest()
    key, chain_code = h[:32], h[32:]
    return key, chain_code


def derive_bip32childkey(parent_key, parent_chain_code, i):
    assert len(parent_key) == 32
    assert len(parent_chain_code) == 32
    k = parent_chain_code
    if (i & BIP32_PRIVDEV) != 0:
        key = b'\x00' + parent_key
    else:
        key = bytes(PublicKey(parent_key))
    d = key + struct.pack('>L', i)
    while True:
        h = hmac.new(k, d, hashlib.sha512).digest()
        key, chain_code = h[:32], h[32:]
        a = int.from_bytes(key, byteorder='big')
        b = int.from_bytes(parent_key, byteorder='big')
        key = (a + b) % BIP32_CURVE.order
        if a < BIP32_CURVE.order and key != 0:
            key = key.to_bytes(32, byteorder='big')
            break
        d = b'\x01' + h[32:] + struct.pack('>L', i)
    return key, chain_code


def parse_derivation_path(str_derivation_path):
    path = []
    if str_derivation_path[0:2] != 'm/':
        raise ValueError("Can't recognize derivation path. It should look like \"m/44'/60/0'/0\".")
    for i in str_derivation_path.lstrip('m/').split('/'):
        if "'" in i:
            path.append(BIP32_PRIVDEV + int(i[:-1]))
        else:
            path.append(int(i))
    return path


def mnemonic_to_private_key(mnemonic, str_derivation_path, passphrase=""):
    derivation_path = parse_derivation_path(str_derivation_path)
    bip39seed = mnemonic_to_bip39seed(mnemonic, passphrase)
    master_private_key, master_chain_code = bip39seed_to_bip32masternode(bip39seed)
    private_key, chain_code = master_private_key, master_chain_code
    for i in derivation_path:
        private_key, chain_code = derive_bip32childkey(private_key, chain_code, i)
    return private_key


if __name__ == '__main__':
    # import sys
    # with open(sys.argv[1], 'r') as f:
    #     mnemonic = ' '.join(f.read().split())

    mnemonic ="elbow hunt follow coin month universe concert lens dish come wrong symbol reason liberty useless famous sport economy talent bundle kit security trumpet uphold"

    private_key = mnemonic_to_private_key(mnemonic,
                                          str_derivation_path=f'{ETH_DERIVATION_PATH}/0')
    public_key = PublicKey(private_key)

    print(f'privkey: {binascii.hexlify(private_key).decode("utf-8")}')
    print(f'pubkey:  {binascii.hexlify(bytes(public_key)).decode("utf-8")}')
    print(f'address: {public_key.address()}')


"""Auth mdoule message types."""

from __future__ import annotations

from glitter_proto.cosmos.vesting.v1beta1 import (
    MsgCreatePeriodicVestingAccount as MsgCreatePeriodicVestingAccount_pb,
)
from glitter_sdk.core.auth.data.periodic_vesting_account import Period

from typing import List
from glitter_sdk.core import AccAddress
from glitter_sdk.core.msg import Msg

__all__ =["MsgCreatePeriodicVestingAccount"]

import attr

@attr.s
class MsgCreatePeriodicVestingAccount(Msg):
    """MsgCreatePeriodicVestingAccount defines a message that enables creating a periodic vesting.
    
    Args:
        from_address (AccAddress): account to create a vesting account
        to_address (AccAddress): vesting account
        start_time (int): vesting start time
        vesting_periods (List[Period]): list of periods
    """

    type_amino = "cosmos-sdk/MsgCreatePeriodicVestingAccount"
    type_url = "/cosmos.vesting.v1beta1.MsgCreatePeriodicVestingAccount"
    prototype = MsgCreatePeriodicVestingAccount_pb

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    start_time: int = attr.ib()
    vesting_periods: List[Period] = attr.ib()

    def to_amino(self) -> dict :
        return {
            "type": self.type_amino,
            "value": {
                "from_address": self.from_address,
                "to_address": self.to_address,
                "start_time": str(self.start_time),
                "vesting_periods": [vp.to_amino() for vp in self.vesting_periods]
            },
        }

    @classmethod
    def from_data(cls, data:dict) -> MsgCreatePeriodicVestingAccount:
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            start_time=int(data["start_time"]),
            vesting_periods=[Period.from_data(vp) for vp in data["vesting_periods"]]
        )
    
    def to_data(self) -> dict :
        return {
            "@type": self.type_url,
            "from_address": self.from_address,
            "to_address": self.to_address,
            "start_time" : str(self.start_time),
            "vesting_periods": [vp.to_data() for vp in self.vesting_periods]
        }

    @classmethod
    def from_proto(cls, proto : MsgCreatePeriodicVestingAccount_pb) -> MsgCreatePeriodicVestingAccount:
        return cls(
            from_address = proto.from_address,
            to_address = proto.to_address,
            start_time = proto.start_time,
            vesting_periods = [Period.from_proto(vs) for vs in proto.vesting_periods]
        )
    
    def to_proto(self) -> MsgCreatePeriodicVestingAccount_pb :
        proto = MsgCreatePeriodicVestingAccount_pb()
        proto.from_address = self.from_address
        proto.to_address = self.to_address
        proto.start_time = self.start_time
        proto.vesting_periods = [vp.to_proto() for vp in self.vesting_periods]
        return proto


from __future__ import annotations
from .raw import RawKey
from glitter_sdk.util.encrypt.eth.ethereum import HDKey
from glitter_sdk.util.encrypt.eth.ethereum import HDPrivateKey

__all__ = ["MnemonicKey", "GLITTER_COIN_TYPE"]

GLITTER_COIN_TYPE = 60


class MnemonicKey(RawKey):
    """A MnemonicKey derives a private key using a BIP39 mnemonic seed phrase, and provides key-derivation options based on the BIP44 HD path standard.

    .. note:: You can change ``coin_type`` to 118 to derive the key for a legacy Glitter
        wallet (shares ``coin_type`` with ATOM).

    Args:
        mnemonic (str, optional): space-separated mnemonic seed phrase. If not provided,
            a 24-word mnemonic will be generated.
        account (int, optional): HD path parameter - account number.
        index (int, optional): HD path parameter - account index.
        coin_type (int, optional): HD path parameter - coin type.
    """

    mnemonic: str
    """Mnemonic key phrase associated with the account (space-separated)."""

    account: int
    """HD path parameter: account number."""

    index: int
    """HD path parameter: account index."""

    coin_type: int
    """HD path parameter: coin type"""

    @property
    def hd_path(self) -> str:
        """Returns the BIP32 HD path for key-derivation:

        ``m/44'/COIN_TYPE'/ACCOUNT'/0/INDEX'``

        Returns:
            str: full BIP32 HD path
        """
        return f"m/44'/{self.coin_type}'/{self.account}'/0/{self.index}"

    def __init__(
        self,
        mnemonic: str = None,
        account: int = 0,
        index: int = 0,
        coin_type: int = GLITTER_COIN_TYPE,
    ):
        master_key = HDPrivateKey.master_key_from_mnemonic(mnemonic)
        root_keys = HDKey.from_path(master_key, "m/44'/60'/0'")
        acct_priv_key = root_keys[-1]
        acct_pub_key = HDKey.from_b58check(acct_priv_key.to_b58check())
        keys = HDKey.from_path(acct_pub_key, f'{0}/{0}')

        # eth_address = keys[-1].public_key.address()
        # self.private_key = bytes.fromhex(keys[-1]._key.to_hex())
        # self.evmos_address = eth_to_evmos(eth_address)
        # self.public_key = keys[-1].public_key.compressed_bytes
        super().__init__(keys[-1])

        self.mnemonic = mnemonic
        self.coin_type = coin_type
        self.account = account
        self.index = index


from __future__ import annotations

import hashlib

from ecdsa import SECP256k1, SigningKey
from ecdsa.util import sigencode_string_canonize

from .key import Key
from glitter_sdk.util.encrypt.eth.ethereum import HDPrivateKey
__all__ = ["RawKey"]

from ..core import PublicKey, SimplePublicKey
from ..core import PublicKey, EthSimplePublicKey
from glitter_sdk.util.encrypt.eth.ethereum import sha3
import coincurve
from coincurve._libsecp256k1 import ffi
from coincurve._libsecp256k1 import lib

class RawKey(Key):
    """RawKey directly uses a raw (plaintext) private key in memory, and provides
    the implementation for signing with ECDSA on curve Secp256k1.

    Args:
        private_key (bytes): private key in bytes
    """

    private_key: bytes
    """Private key, in bytes."""

    @classmethod
    def from_hex(cls, private_key_hex: str) -> RawKey:
        """Create a new RawKey from a hex-encoded private key string.

        Args:
            private_key_hex (str): hex-encoded private key
        """
        return cls(bytes.fromhex(private_key_hex))

    def __init__(self, private_key: HDPrivateKey):
        public_key = private_key.public_key.compressed_bytes
        super().__init__(EthSimplePublicKey(key=public_key))
        self.private_key = bytes.fromhex(private_key._key.to_hex())


    def sign(self, payload: bytes) -> bytes:
        """Signs the data payload using ECDSA and curve Secp256k1 with the private key as
        the signing key.

        Args:
            payload (bytes): data to sign
        """
        digest = sha3(payload)
        return self.evmSign(digest)

    def evmSign(self, msg: bytes) -> bytes:
        key = coincurve.PrivateKey(self.private_key)
        nonce = (lib.secp256k1_nonce_function_rfc6979, ffi.NULL)
        return key.sign_recoverable(msg, hasher=None, custom_nonce=nonce)



import binascii
from typing import Union

from glitter_sdk.util.encrypt.converter.bech32 import bech32_decode
from glitter_sdk.util.encrypt.converter.bech32 import bech32_encode
from glitter_sdk.util.encrypt.converter.bech32 import convertbits
from glitter_sdk.util.encrypt.converter.bech32 import Encoding

EVMOS_PREFIX = 'glitter'
ETH_PREFIX = '0x'


def bech32_to_eth(wallet: str, prefix: str) -> Union[str, None]:
    decoded = bech32_decode(wallet)
    if decoded[0] != prefix:
        return None
    words = convertbits(decoded[1], 5, 8, False)
    res = ''
    for w in words:
        res = f'{res}{format(w, "x").zfill(2)}'

    return f'{ETH_PREFIX}{res}'


def eth_to_bech32(wallet: str, prefix: str) -> str:
    splitted = wallet.split('0x')
    if len(splitted) == 2:
        raw_address = splitted[1]
    else:
        raw_address = wallet
    try:
        array = binascii.unhexlify(raw_address)
        words = [x for x in array]
        bech32_words = convertbits(words, 8, 5)
        bech32_address = bech32_encode(prefix, bech32_words, Encoding.BECH32)
    except Exception:
        return None
    if len(bech32_address) != 46:
        return None

    return bech32_address


def glitter_to_eth(wallet: str) -> Union[str, None]:
    return bech32_to_eth(wallet, EVMOS_PREFIX)


def eth_to_glitter(wallet: str) -> str:
    return eth_to_bech32(wallet, EVMOS_PREFIX)



"""Bank module message types."""

from __future__ import annotations

from typing import List

from betterproto.lib.google.protobuf import Any as Any_pb
from glitter_proto.cosmos.bank.v1beta1 import Input as Input_pb
from glitter_proto.cosmos.bank.v1beta1 import MsgMultiSend as MsgMultiSend_pb
from glitter_proto.blockved.glitterchain.index import SqlExecRequest as SQLExecRequest_pb
from glitter_proto.blockved.glitterchain.index import SqlGrantRequest as SQLGrantRequest_pb
from glitter_proto.cosmos.bank.v1beta1 import Output as Output_pb

from glitter_sdk.core import AccAddress
from glitter_sdk.core.msg import Msg
from glitter_sdk.util.json import JSONSerializable
from .arguments import Arguments

__all__ = ["SQLExecRequest", "SQLGrantRequest"]

import attr


@attr.s
class SQLExecRequest(Msg):
    """Sends sql request.

    Args:
        uid (AccAddress): sender
        sql (AccAddress): recipient
        arguments (Arguments): arguments to send
    """

    type_amino = "cosmos-sdk/SQLExecRequest"
    """"""
    type_url = "/blockved.glitterchain.index.SQLExecRequest"
    """"""
    action = "send"
    """"""
    prototype = SQLExecRequest_pb
    """"""

    uid: AccAddress = attr.ib()
    sql: AccAddress = attr.ib()
    arguments: Arguments = attr.ib(converter=Arguments,default=None)

    @classmethod
    def from_proto(cls, proto: SQLExecRequest_pb) -> SQLExecRequest:
        return cls(
            uid=proto.uid,
            sql=proto.sql,
            arguments=Arguments.from_proto(proto.arguments)
        )

    def to_proto(self) -> SQLExecRequest_pb:
        proto = SQLExecRequest_pb()
        proto.uid = self.uid
        proto.sql = self.sql
        proto.arguments = [c.to_proto() for c in self.arguments]
        return proto


@attr.s
class SQLGrantRequest(Msg):
    """Grant ``role`` to ``to_u_i_d`` .

    Args:
        uid (AccAddress): sender
        to_uid (AccAddress): recipient
        role (str): role
        on_database (str): database
        on_table (str): table
    """

    type_amino = "cosmos-sdk/SQLGrantRequest"
    """"""
    type_url = "/blockved.glitterchain.index.SQLGrantRequest"
    """"""
    action = "send"
    """"""
    prototype = SQLGrantRequest_pb
    """"""

    uid: AccAddress = attr.ib()
    to_uid: AccAddress = attr.ib()
    role: AccAddress = attr.ib()
    on_database: AccAddress = attr.ib()
    on_table: AccAddress = attr.ib(default=None)

    @classmethod
    def from_proto(cls, proto: SQLGrantRequest_pb) -> SQLGrantRequest:
        return cls(
            uid=proto.uid,
            to_uid=proto.to_uid,
            role=proto.role,
            on_database=proto.on_database,
            on_table=proto.on_table,
        )

    def to_proto(self) -> SQLGrantRequest_pb:
        proto = SQLGrantRequest_pb()
        proto.uid = self.uid
        proto.to_uid = self.to_uid
        proto.role = self.role
        proto.on_database = self.on_database
        proto.on_table = self.on_table
        return proto



"""Auth mdoule message types."""

from __future__ import annotations

from glitter_proto.cosmos.vesting.v1beta1 import (
    MsgCreateVestingAccount as MsgCreateVestingAccount_pb,
)

from typing import List
from glitter_sdk.core import AccAddress, Coins
from glitter_sdk.core.msg import Msg

__all__ =["MsgCreateVestingAccount"]

import attr

@attr.s
class MsgCreateVestingAccount(Msg):
    """MsgCreateVestingAccount defines a message that enables creating a vesting.
    
    Args:
        from_address (AccAddress): account to create a vesting account
        to_address (AccAddress): vesting account
        amount (Coins): vesting amount
        end_time (int): vesting end time
        delayed (bool): all coins are vested once end time is reached 
    """

    type_amino = "cosmos-sdk/MsgCreateVestingAccount"
    type_url = "/cosmos.vesting.v1beta1.MsgCreateVestingAccount"
    prototype = MsgCreateVestingAccount_pb

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)
    end_time: int = attr.ib()
    delayed: bool = attr.ib()

    def to_amino(self) -> dict :
        return {
            "type": self.type_amino,
            "value": {
                "from_address": self.from_address,
                "to_address": self.to_address,
                "amount": self.amount.to_amino(),
                "end_time" : str(self.end_time),
                "delayed" : self.delayed
            },
        }

    @classmethod
    def from_data(cls, data:dict) -> MsgCreateVestingAccount:
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            amount=Coins.from_data(data["amount"]),
            end_time=int(data["end_time"]),
            delayed=data["delayed"]
        )
    
    def to_data(self) -> dict :
        return {
            "@type": self.type_url,
            "from_address": self.from_address,
            "to_address": self.to_address,
            "amount": self.amount.to_data(),
            "end_time": str(self.end_time),
            "delayed" : self.delayed
        }

    @classmethod
    def from_proto(cls, proto : MsgCreateVestingAccount_pb) -> MsgCreateVestingAccount:
        return cls(
            from_address = proto.from_address,
            to_address = proto.to_address,
            amount = Coins.from_proto(proto.amount),
            end_time = proto.end_time,
            delayed = proto.delayed
        )
    
    def to_proto(self) -> MsgCreateVestingAccount_pb :
        proto = MsgCreateVestingAccount_pb()
        proto.from_address = self.from_address
        proto.to_address = self.to_address
        proto.amount = [c.to_proto() for c in self.amount]
        proto.end_time = self.end_time
        proto.delayed = self.delayed
        return proto


from attrs import asdict, define, make_class, Factory

_reserved_chars = {}
_reserved_chars[ord("+")] = "\\+"
_reserved_chars[ord("-")] = "\\-"
_reserved_chars[ord("=")] = "\\="
_reserved_chars[ord("&")] = "\\&"
_reserved_chars[ord("|")] = "\\|"
_reserved_chars[ord(">")] = "\\>"
_reserved_chars[ord("<")] = "\\<"
_reserved_chars[ord("!")] = "\\!"
_reserved_chars[ord("(")] = "\\("
_reserved_chars[ord(")")] = "\\)"
_reserved_chars[ord("{")] = "\\{"
_reserved_chars[ord("}")] = "\\}"
_reserved_chars[ord("[")] = "\\["
_reserved_chars[ord("]")] = "\\]"
_reserved_chars[ord("^")] = "\\^"
_reserved_chars[ord('"')] = "\\"
_reserved_chars[ord("~")] = "\\~"
_reserved_chars[ord("*")] = "\\*"
_reserved_chars[ord("?")] = "\\?"
_reserved_chars[ord(":")] = "\\:"
_reserved_chars[ord("\\")] = "\\\\"
_reserved_chars[ord("/")] = "\/"
_reserved_chars[ord(" ")] = "\ "

_escape_char = {}
_escape_char[ord("\\")] = "\\\\"


def reserve_query(value, mapping=None):
    """escapes *value* without adding quote.

    Value should be unicode
    """
    return value.translate(_reserved_chars)


def escape_query(value, mapping=None):
    """escapes *value* without adding quote.

    Value should be unicode
    """
    return value.translate(_escape_char)


def query_string_prepare(queries: list):
    query_string = []
    for query in queries:
        query_string.append(str(query))

    rst = " ".join(query_string)
    return rst


@define
class Query:
    def prepare_query(self):
        pass


@define
class MatchQuery:
    field: str
    query: str
    boost: float = 1.0

    def prepare_query(self):
        rst = "{}:{}^{}".format(self.field, reserve_query(self.query), self.boost)
        return rst

    def __str__(self):
        return self.prepare_query()


@define
class MatchPhraseQuery:
    field: str
    query: str
    boost: float = 1.0

    def prepare_query(self):
        return "{}:{}^{}".format(self.field, "\"" + self.query + "\"", self.boost)

    def __str__(self):
        return self.prepare_query()


@define
class RegexpQuery:
    field: str
    query: str
    boost: float = 1.0

    def prepare_query(self):
        return "{}:{}^{}".format(self.field, "/" + self.query + "/", self.boost)

    def __str__(self):
        return self.prepare_query()


@define
class NumericRangeQuery:
    field: str
    operator: str
    value: int
    boost: float = 1.0

    def prepare_query(self):
        return "{}:{}{}^{}".format(self.field, self.operator, self.value, self.boost)

    def __str__(self):
        return self.prepare_query()


@define
class DateRangeQuery:
    field: str
    operator: str
    value: str
    boost: float = 1.0

    def prepare_query(self):
        return "{}:{}\"{}\"^{}".format(self.field, self.operator, self.value, self.boost)

    def __str__(self):
        return self.prepare_query()



try:
    from glitter_sdk.util.encrypt.crypto import ecdsa_openssl as _ecdsa
except Exception:
    from glitter_sdk.util.encrypt.crypto import ecdsa_python as _ecdsa

ECPointAffine = _ecdsa.ECPointAffine
EllipticCurve = _ecdsa.EllipticCurve
secp256k1 = _ecdsa.secp256k1



"""This submodule provides functions for accomplishing common tasks encountered
in creating and parsing Bitcoin objects, like turning difficulties into targets
or deserializing and serializing various kinds of packed byte formats."""
import codecs
import hashlib
import os
import random
import struct

import base58

MAX_TARGET = 0x00000000FFFF0000000000000000000000000000000000000000000000000000


def rand_bytes(n, secure=True):
    """ Returns n random bytes.

    Args:
        n (int): number of bytes to return.
        secure (bool): If True, uses os.urandom to generate
            cryptographically secure random bytes. Otherwise, uses
            random.randint() which generates pseudo-random numbers.

    Returns:
        b (bytes): n random bytes.
    """
    if secure:
        return os.urandom(n)
    else:
        return bytes(random.randint(0, 255) for i in range(n))


def bytes_to_str(b):
    """ Converts bytes into a hex-encoded string.

    Args:
        b (bytes): bytes to encode

    Returns:
        h (str): hex-encoded string corresponding to b.
    """
    return codecs.encode(b, 'hex_codec').decode('ascii')


def hex_str_to_bytes(h):
    """ Converts a hex-encoded string to bytes.

    Args:
        h (str): hex-encoded string to convert.

    Returns:
        b (bytes): bytes corresponding to h.
    """
    return bytes.fromhex(h)


# Is there a better way of doing this?
def render_int(n):
    """ Renders an int in the shortest possible form.

    When packing the height into the coinbase script, the integer
    representing the height must be encoded in the shortest possible
    manner. See: https://bitcoin.org/en/developer-reference#coinbase.

    Args:
        n (int): number to be encoded.

    Returns:
        b (bytes): bytes representing n in the shortest possible form.
    """
    # little-endian byte stream
    if n < 0:
        neg = True
        n = -n
    else:
        neg = False
    r = []
    while n:
        r.append(n & 0xff)
        n >>= 8
    if neg:
        if r[-1] & 0x80:
            r.append(0x80)
        else:
            r[-1] |= 0x80
    elif r and (r[-1] & 0x80):
        r.append(0)
    return bytes(r)


def pack_compact_int(i):
    """ See
    https://bitcoin.org/en/developer-reference#compactsize-unsigned-integers

    Args:
        i (int): Integer to be serialized.

    Returns:
        b (bytes): Serialized bytes corresponding to i.
    """
    if i < 0xfd:
        return struct.pack('<B', i)
    elif i <= 0xffff:
        return struct.pack('<BH', 0xfd, i)
    elif i <= 0xffffffff:
        return struct.pack('<BI', 0xfe, i)
    else:
        return struct.pack('<BQ', 0xff, i)


def unpack_compact_int(bytestr):
    """ See
    https://bitcoin.org/en/developer-reference#compactsize-unsigned-integers

    Args:
        bytestr (bytes): bytes containing an unsigned integer to be
            deserialized.

    Returns:
        n (int): deserialized integer.
    """

    b0 = bytestr[0]
    if b0 < 0xfd:
        return (b0, bytestr[1:])
    elif b0 == 0xfd:
        return (struct.unpack('<H', bytestr[1:3])[0], bytestr[3:])
    elif b0 == 0xfe:
        return (struct.unpack('<I', bytestr[1:5])[0], bytestr[5:])
    elif b0 == 0xff:
        return (struct.unpack('<Q', bytestr[1:9])[0], bytestr[9:])
    else:
        return None


def pack_u32(i):
    """ Serializes a 32-bit integer into little-endian form.

    Args:
        i (int): integer to be serialized.

    Returns:
        b (bytes): 4 bytes containing the little-endian serialization of i.
    """
    return struct.pack('<I', i)


def unpack_u32(b):
    """ Deserializes a 32-bit integer from bytes.

    Args:
        b (bytes): At least 4 bytes containing the serialized integer.

    Returns:
        (i, b) (tuple): A tuple containing the deserialized integer and the
        remainder of the byte stream.
    """
    u32 = struct.unpack('<I', b[0:4])
    return (u32[0], b[4:])


def pack_u64(i):
    """ Serializes a 64-bit integer into little-endian form.

    Args:
        i (int): integer to be serialized.

    Returns:
        b (bytes): 8 bytes containing the little-endian serialization of i.
    """
    return struct.pack('<Q', i)


def unpack_u64(b):
    """ Deserializes a 64-bit integer from bytes.

    Args:
        b (bytes): At least 8 bytes containing the serialized integer.

    Returns:
        (i, b) (tuple): A tuple containing the deserialized integer and the
        remainder of the byte stream.
    """
    u64 = struct.unpack('<Q', b[0:8])
    return (u64[0], b[8:])


def pack_var_str(s):
    """ Serializes a variable length byte stream.

    Args:
        s (bytes): byte stream to serialize

    Return:
        b (bytes): Serialized bytes, prepended with the length of the
        byte stream.
    """
    return pack_compact_int(len(s)) + s


def unpack_var_str(b):
    """ Deserializes a variable length byte stream.

    Args:
        b (bytes): variable length byte stream to deserialize

    Returns:
        (s, b) (tuple): A tuple containing the variable length byte stream
        and the remainder of the input byte stream.
    """
    strlen, b0 = unpack_compact_int(b)
    return (b0[:strlen], b0[strlen:])


def bits_to_target(bits):
    """ Decodes the full target from a compact representation.
    See: https://bitcoin.org/en/developer-reference#target-nbits

    Args:
        bits (int): Compact target (32 bits)

    Returns:
        target (Bignum): Full 256-bit target
    """
    shift = bits >> 24
    target = (bits & 0xffffff) * (1 << (8 * (shift - 3)))
    return target


def bits_to_difficulty(bits):
    """ Determines the difficulty corresponding to bits.
    See: https://en.bitcoin.it/wiki/Difficulty

    Args:
        bits (int): Compact target (32 bits)

    Returns:
        diff (float): Measure of how hard it is to find a solution
        below the target represented by bits.
    """
    target = bits_to_target(bits)
    return MAX_TARGET / target


def difficulty_to_target(difficulty):
    """ Converts a difficulty to a long-form target.

    Args:
        difficulty (float): The difficulty to return the appropriate target for

    Returns:
        target (int): The corresponding target
    """
    return int(MAX_TARGET / difficulty)


def target_to_bits(target):
    """ Creates a compact target representation for a given target.

    Args:
        target (Bignum): The long-form target to make compact.

    Returns:
        ct (int): Compact target
    """
    # Get bit length
    nbits = target.bit_length()
    # Round up to next 8-bits
    nbits = ((nbits + 7) & ~0x7)
    exponent = (int(nbits / 8) & 0xff)
    coefficient = (target >> (nbits - 24)) & 0xffffff
    if coefficient & 0x800000:
        coefficient >>= 8
        exponent += 1
    return (exponent << 24) | coefficient


def difficulty_to_bits(difficulty):
    """ Converts a difficulty to a compact target.

    Args:
        difficulty (float): The difficulty to create a target for

    Returns:
        ct (int): Compact target
    """
    return target_to_bits(difficulty_to_target(difficulty))


def address_to_key_hash(s):
    """ Given a Bitcoin address decodes the version and
    RIPEMD-160 hash of the public key.

    Args:
        s (bytes): The Bitcoin address to decode

    Returns:
        (version, h160) (tuple): A tuple containing the version and
        RIPEMD-160 hash of the public key.
    """
    n = base58.b58decode_check(s)
    version = n[0]
    h160 = n[1:]
    return version, h160


def key_hash_to_address(hash160, version=0x0):
    """Convert RIPEMD-160 hash to bitcoin address.

    Args:
        hash160 (bytes/str): bitcoin hash160 to decode
        version (int): The version prefix

    Returns:
        (bitcoin address): base58 encoded bitcoin address
    """
    if isinstance(hash160, str):
        # if 0x in string, strip it
        if '0x' in hash160:
            h160 = hex_str_to_bytes(hash160[2:])
        else:
            h160 = hex_str_to_bytes(hash160)
    elif isinstance(hash160, bytes):
        h160 = hash160

    address = base58.b58encode_check(bytes([version]) + h160)
    return address


def hash160(b):
    """ Computes the HASH160 of b.

    Args:
        b (bytes): A byte string to compute the HASH160 of.

    Returns:
        The RIPEMD-160 digest of the SHA256 hash of b.
    """
    r = hashlib.new('ripemd160')
    r.update(hashlib.sha256(b).digest())

    return r.digest()


def compute_reward(height):
    """ Computes the block reward for a block at the supplied height.
    See: https://en.bitcoin.it/wiki/Controlled_supply for the reward
    schedule.

    Args:
        height (int): Block height

    Returns:
        reward (int): Number of satoshis rewarded for solving a block at the
        given height.
    """
    base_subsidy = 50 * 100000000
    era = height // 210000
    if era == 0:
        return base_subsidy
    return int(base_subsidy / 2**era)



from .msg_create_periodic_vesting_account import MsgCreatePeriodicVestingAccount
from .msg_create_vesting_account import MsgCreateVestingAccount
from .msg_donate_all_vesting_tokens import MsgDonateAllVestingTokens

__all__ = [
    "MsgCreatePeriodicVestingAccount",
    "MsgCreateVestingAccount",
    "MsgDonateAllVestingTokens"
]



from __future__ import annotations

import requests
import base64
import glitter_proto.blockved.glitterchain.index
from glitter_sdk.key.key import Key, SignOptions
from glitter_sdk.core.tx import SignMode, Tx
from glitter_sdk.core.msgs import SQLExecRequest, SQLGrantRequest
from glitter_sdk.util.url import urljoin
from glitter_sdk.util.parse_sql import escape_args
from glitter_sdk.exceptions import *
from glitter_sdk.client.lcd.api.tx import CreateTxOptions, SignerOptions
from glitter_proto.blockved.glitterchain.index import *
from glitter_sdk.core.bank import MsgMultiSend, MsgSend

__all__ = ["DB", "AsyncDB"]

GrantReader = "reader"
GrantWriter = "writer"
GrantAdmin = "admin"
WriteLimitRow = 50


class AsyncDB:
    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    async def account_number(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    async def sequence(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    async def account_number_and_sequence(self) -> dict:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.get_account_number(), "sequence": res.get_sequence()}

    async def create_tx(self, options: CreateTxOptions) -> Tx:
        sigOpt = [
            SignerOptions(
                address=self.key.acc_address,
                sequence=options.sequence,
                public_key=self.key.public_key,
            )
        ]
        return await self.lcd.tx.create(sigOpt, options)

    async def create_and_sign_tx(self, options: CreateTxOptions) -> Tx:
        account_number = options.account_number
        sequence = options.sequence
        if account_number is None or sequence is None:
            res = await self.account_number_and_sequence()
            if account_number is None:
                account_number = res.get("account_number")
            if sequence is None:
                sequence = res.get("sequence")
        options.sequence = sequence
        options.account_number = account_number
        return self.key.sign_tx(
            tx=(await self.create_tx(options)),
            options=SignOptions(
                account_number=account_number,
                sequence=sequence,
                chain_id=self.lcd.chain_id,
                sign_mode=options.sign_mode
                if options.sign_mode
                else SignMode.SIGN_MODE_DIRECT,
            ),
        )


class DB:
    """Wraps around a :class:`Key` implementation and provides transaction building and
    signing functionality. It is recommended to create this object through
    :meth:`LCDClient.DB()<glitter_sdk.client.lcd.LCDClient.DB>`."""

    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    def account_number(self) -> int:
        """Fetches account number for the account associated with the Key."""
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    def sequence(self) -> int:
        """Fetches the sequence number for the account associated with the Key."""
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    def account_number_and_sequence(self) -> dict:
        """Fetches both account and sequence number associated with the Key."""
        k = self.key.acc_address
        res = self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.get_account_number(), "sequence": res.get_sequence()}

    def create_tx(self, options: CreateTxOptions) -> Tx:
        """Builds an unsigned transaction object. The ``DB`` will first
        query the blockchain to fetch the latest ``account`` and ``sequence`` values for the
        account corresponding to its Key, unless the they are both provided. If no ``fee``
        parameter is set, automatic fee estimation will be used (see `fee_estimation`).

        Args:
            options (CreateTxOptions): Options to create a tx

        Returns:
            Tx: unsigned transaction
        """
        sigOpt = [
            SignerOptions(
                address=self.key.acc_address,
                sequence=options.sequence,
                public_key=self.key.public_key,
            )
        ]
        return self.lcd.tx.create(sigOpt, options)

    def create_and_sign_tx(self, options: CreateTxOptions) -> Tx:
        """Creates and signs a :class:`Tx` object in a single step. This is the recommended
        method for preparing transaction for immediate signing and broadcastring. The transaction
        is generated exactly as :meth:`create_tx`.

        Args:
            options (CreateTxOptions): Options to create a tx

        Returns:
            Tx: signed transaction
        """

        account_number = options.account_number
        sequence = options.sequence
        if account_number is None or sequence is None:
            res = self.account_number_and_sequence()
            if account_number is None:
                account_number = res.get("account_number")
            if sequence is None:
                sequence = res.get("sequence")
        options.sequence = sequence
        options.account_number = account_number
        return self.key.sign_tx(
            tx=self.create_tx(options),
            options=SignOptions(
                account_number=account_number,
                sequence=sequence,
                chain_id=self.lcd.chain_id,
                sign_mode=options.sign_mode
                if options.sign_mode
                else SignMode.SIGN_MODE_DIRECT,
            ),
        )

    def sql_exec(self, sql):
        tx = self.create_and_sign_tx(CreateTxOptions(
            msgs=[SQLExecRequest(
                self.key.acc_address,
                sql,
            )],
            # fee=Fee(300000, "15stake"),
            memo="sql transaction!",
            fee_denoms=["agli"],
            sign_mode=SignMode.SIGN_MODE_DIRECT,
            gas="auto",
        ))
        return self.lcd.tx.broadcast(tx)

    def create_database(self, database):
        sql = "CREATE DATABASE IF NOT EXISTS {} ".format(database)
        return self.sql_exec(sql)

    def create_table(self, sql):
        return self.sql_exec(sql)

    def drop_table(self, database: str, table: str):
        sql = "DROP TABLE IF EXISTS {}.{} ".format(database, table)
        return self.sql_exec(sql)

    # only suport standard engine
    def add_column(self, database: str, table: str, col: str):
        sql = "ALTER TABLE {}.{} ADD {} ".format(database, table, col)
        return self.sql_exec(sql)

    def show_create_table(self, database: str, table: str):
        endpoint = "/blockved/glitterchain/index/sql/show_create_table"
        req = ShowCreateTableRequest()
        req.database_name = database
        req.table_name = table
        r = requests.get(urljoin(self.lcd.url, endpoint), req.to_dict())
        if r.status_code != 200:
            raise LCDResponseError(message=r.text, response=r)
        response = ShowCreateTableResponse().from_dict(r.json())
        return response

    def list_databases(self, creator: str = None):
        endpoint = "/blockved/glitterchain/index/sql/list_databases"
        r = requests.get(urljoin(self.lcd.url, endpoint))
        if r.status_code != 200:
            raise LCDResponseError(message=str(r.status_code), response=r)

        result = SqlListDatabasesResponse()
        response = SqlListDatabasesResponse().from_dict(r.json())
        for  db in response.databases:
            if creator and creator != db.creator:
                continue
            result.databases.append(db)
        return result

    def list_tables(self, table_keyword: str = None, uid: str = None, database: str = None,
                    page: int = None, page_size: int = None):
        endpoint = "/blockved/glitterchain/index/sql/list_tables"

        payload = {}
        if table_keyword is not None:
            payload["keyword"] = table_keyword
        if uid is not None:
            payload["uid"] = uid
        if database is not None:
            payload["database"] = database
        if page is not None:
            payload["page"] = page
        if page_size is not None:
            payload["page_size"] = page_size

        r = requests.get(urljoin(self.lcd.url, endpoint), payload)
        if r.status_code != 200:
            raise LCDResponseError(message=str(r.status_code), response=r)
        return SqlListTablesResponse().from_dict(r.json())

    def insert(self, table: str, columns: map):
        col_name = list(columns.keys())
        place_hold = []
        vals = []
        for idx in col_name:
            place_hold.append('%s')
            vals.append(columns[idx])
        vals = escape_args(vals)

        sql = "insert into {} ({}) values ({})".format(table, ",".join(col_name), ",".join(place_hold))
        stmt = sql % vals
        return self.sql_exec(stmt)

    def batch_insert(self, db: str, table: str, rows: list = None):
        if not rows:
            raise ParamError("rows is empty")

        col_names = rows[0].keys()
        row_len = len(rows)
        col_len = len(col_names)
        place_hold = []
        vals = []
        for row_idx in range(0, row_len):
            place_hold_arr = ['%s'] * col_len
            place_hold.append("(" + ",".join(place_hold_arr) + ")")

        for row in rows:
            for col_name in col_names:
                vals.append(row[col_name])
        vals = escape_args(vals)

        sql = "INSERT INTO {}.{} ({}) VALUES {}".format(db, table, ",".join(col_names),
                                                        ",".join(place_hold))
        stmt = sql % vals
        return self.sql_exec(stmt)

    def update(self, database_name: str, table_name: str, columns: map = None, where: map = None):
        if not columns:
            raise ParamError("columns is empty")

        update = []
        for col_name, col_val in columns.items():
            update.append("{}={}".format(col_name, escape_args(col_val)))

        sql = "UPDATE  {}.{} SET {} ".format(database_name, table_name, ",".join(update))

        where_cond = []
        if where:
            for col_name, col_val in where.items():
                where_cond.append("{}{}".format(col_name, escape_args(col_val)))
            sql += " WHERE {}".format(" AND ".join(where_cond))

        return self.sql_exec(sql)

    def delete(self, database_name: str, table_name: str, where: map, order_by: str = None,
               asc: bool = True, limit: int = WriteLimitRow):
        if not where:
            raise ParamError("where is empty")
        if limit > WriteLimitRow:
            raise ParamError("too much will to delete")

        sql = "DELETE FROM {}.{} ".format(database_name, table_name)

        where_cond = []
        if where:
            for col_name, col_val in where.items():
                where_cond.append("{}{}".format(col_name, escape_args(col_val)))
            sql += " WHERE {}".format(" AND ".join(where_cond))

        if order_by:
            sql += " ORDER BY {} {}".format(order_by, "ASC" if asc else "DESC")

        sql += " LIMIT {}".format(limit)

        return self.sql_exec(sql)

    def query(self, sql: str):
        endpoint = "/blockved/glitterchain/index/sql/simple_query"
        req = SqlQueryRequest()
        req.sql = sql
        r = requests.post(urljoin(self.lcd.url, endpoint), data=req.to_json())
        if r.status_code != 200:
            raise LCDResponseError(message=r.text, response=str(r.status_code))
        response = SimpleSqlQueryResponse().from_dict(r.json())
        row_set = []
        for raw_row in response.result:
            row = {}
            for field_name, col_val in raw_row.row.items():
                value_type = col_val.column_value_type
                if value_type == ColumnValueType.IntColumn or value_type == ColumnValueType.UintColumn:
                    row[field_name] = int(col_val.value)
                elif value_type == ColumnValueType.FloatColumn:
                    row[field_name] = float(col_val.value)
                elif value_type == ColumnValueType.BoolColumn:
                    row[field_name] = bool(col_val.value)
                elif value_type == ColumnValueType.StringColumn:
                    row[field_name] = col_val.value
                elif value_type == ColumnValueType.BytesColumn:
                    row[field_name] = base64.b64decode(col_val.value)
                elif value_type == ColumnValueType.InvalidColumn:
                    row[field_name] = col_val.value
            row_set.append(row)
        return row_set

    def sql_grant(self, to_addr: str, role: str, database: str, table: str = None):
        grantRequest = SQLGrantRequest(
            self.key.acc_address,
            to_addr,
            role,
            database
        )
        if table:
            grantRequest.on_table = table

        tx = self.create_and_sign_tx(CreateTxOptions(
            msgs=[grantRequest],
            # fee=Fee(300000, "15agli"),
            memo="grant transaction!",
            fee_denoms=["agli"],
            sign_mode=SignMode.SIGN_MODE_DIRECT
        ))
        return self.lcd.tx.broadcast(tx)

    def grant_reader(self, to_addr: str, database: str, table: str = None):
        return self.sql_grant(to_addr, GrantReader, database, table)

    def grant_writer(self, to_addr: str, database: str, table: str = None):
        return self.sql_grant(to_addr, GrantWriter, database, table)

    def grant_admin(self, to_addr: str, database: str, table: str = None):
        return self.sql_grant(to_addr, GrantAdmin, database, table)

    def transfer(self, addr: str, amount: str):
        tx = self.create_and_sign_tx(CreateTxOptions(
            msgs=[MsgSend(
                self.key.acc_address,
                addr,
                amount
            )],
            # fee=Fee(300000, "15stake"),
            memo="ban send transaction!",
            fee_denoms=["agli"],
            sign_mode=SignMode.SIGN_MODE_DIRECT,
            gas="auto",
        ))
        return self.lcd.tx.broadcast(tx)



from glitter_sdk.core.public_key import PublicKey
from glitter_sdk.core.tx import TxInfo, TxLog, parse_tx_logs

from .account import Account
from .base_account import BaseAccount
from .continuous_vesting_account import ContinuousVestingAccount
from .delayed_vesting_account import DelayedVestingAccount
from .periodic_vesting_account import (
    PeriodicVestingAccount,
    Period
)
from .eth_account import EthAccount

__all__ = [
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
    "Account",
    "BaseAccount",
    "ContinuousVestingAccount",
    "DelayedVestingAccount",
    "PeriodicVestingAccount",
    "Period",
    "EthAccount"
]



import coincurve
from coincurve._libsecp256k1 import ffi
from coincurve._libsecp256k1 import lib

from glitter_sdk.util.encrypt.converter import eth_to_glitter
from glitter_sdk.util.encrypt.eth.ethereum import HDKey
from glitter_sdk.util.encrypt.eth.ethereum import HDPrivateKey


class DB:

    def __init__(self, seed, algo='ethsecp256k1') -> None:
        if algo == 'ethsecp256k1':
            master_key = HDPrivateKey.master_key_from_mnemonic(seed)
            root_keys = HDKey.from_path(master_key, "m/44'/60'/0'")

            acct_priv_key = root_keys[-1]

            acct_pub_key = HDKey.from_b58check(acct_priv_key.to_b58check())
            keys = HDKey.from_path(acct_pub_key, f'{0}/{0}')

            self.eth_address = keys[-1].public_key.address()
            self.private_key = bytes.fromhex(keys[-1]._key.to_hex())
            self.evmos_address = eth_to_glitter(self.eth_address)
            self.public_key = keys[-1].public_key.compressed_bytes
        else:
            # TODO: sopport for secp256k1
            raise NotImplementedError(f'{algo} is not supported!')

    def sign(self, msg: bytes) -> bytes:
        key = coincurve.PrivateKey(self.private_key)
        nonce = (lib.secp256k1_nonce_function_rfc6979, ffi.NULL)
        return key.sign_recoverable(msg, hasher=None, custom_nonce=nonce)
